"""This module is provide internal util functions."""

import collections
import functools
import json
import ssl
from os import path
from typing import Mapping, Optional, Union
from urllib import request

from mimesis import config
from mimesis.exceptions import UnsupportedLocale
from mimesis.typing import JSON

__all__ = ['download_image', 'luhn_checksum',
           'setup_locale', 'pull']

DATA_DIR = path.abspath(path.join(path.dirname(__file__), 'data'))


def luhn_checksum(num: str) -> str:
    """Calculate a checksum for num using the Luhn algorithm.

    :param num: The number to calculate a checksum for as a string.
    :return: Checksum for number.
    """
    check = 0
    for i, s in enumerate(reversed(num)):
        sx = int(s)
        sx = sx * 2 if i % 2 == 0 else sx
        sx = sx - 9 if sx > 9 else sx
        check += sx
    return str(check * 9 % 10)


def update_dict(initial: JSON, other: Mapping) -> JSON:
    """Recursively update a dictionary.

    :param initial: Dict to update.
    :type initial: dict or list
    :param other: Dict to update from.
    :type other: Mapping
    :return: Updated dict.
    :rtype: dict
    """
    for key, value in other.items():
        if isinstance(value, collections.Mapping):
            r = update_dict(initial.get(key, {}), value)
            initial[key] = r
        else:
            initial[key] = other[key]
    return initial


@functools.lru_cache(maxsize=None)
def pull(file: str, locale: str = 'en') -> JSON:
    """Pull the content from the JSON and memorize one.

    Opens JSON file ``file`` in the folder ``data/locale``
    and get content from the file and memorize ones using lru_cache.

    :param file: The name of file.
    :param locale: Locale.
    :return: The content of the file.
    :rtype: dict
    :raises UnsupportedLocale: if locale is not supported.

    :Example:

        >>> from mimesis.utils import pull
        >>> en = pull(file='datetime.json', locale='en')
        >>> isinstance(en, dict)
        True
        >>> en['day']['abbr'][0]
        'Mon.'
    """
    def get_data(locale_name: str) -> JSON:
        """Pull JSON data from file.

        :param locale_name: Locale name.
        :return: Content of JSON file as dict.
        """
        file_path = path.join(DATA_DIR, locale_name, file)
        with open(file_path, 'r', encoding='utf8') as f:
            return json.load(f)

    locale = locale.lower()

    if locale not in config.SUPPORTED_LOCALES:
        raise UnsupportedLocale(locale)

    master_locale = locale.split('-')[0]
    data = get_data(master_locale)

    # Handle sub-locales
    if '-' in locale:
        data = update_dict(data, get_data(locale))

    return data


def download_image(url: str = '', save_path: str = '',
                   unverified_ctx: bool = False) -> Union[None, str]:
    """Download image and save in current directory on local machine.

    :param url: URL to image.
    :param save_path: Saving path.
    :param unverified_ctx: Create unverified context.
    :return: Image name.
    :rtype: str or None
    """
    if unverified_ctx:
        ssl._create_default_https_context = ssl._create_unverified_context

    if url:
        image_name = url.rsplit('/')[-1]
        request.urlretrieve(url, save_path + image_name)
        return image_name
    return None


def setup_locale(locale: Optional[str] = None) -> str:
    """Set up locale after pre-check.

    :param str locale: Locale
    :return: Locale in lowercase.
    :raises UnsupportedLocale: if locales is not supported.
    """
    if not locale:
        return config.DEFAULT_LOCALE

    locale = locale.lower()
    if locale not in config.SUPPORTED_LOCALES:
        raise UnsupportedLocale(locale)

    return locale



import json
import os
from distutils.core import setup
from os.path import abspath, dirname, exists, getsize, join, relpath, splitext

from setuptools import Command

here = abspath(dirname(__file__))

about = {}
# Get meta-data from __version__.py
with open(join(here, 'mimesis', '__version__.py')) as f:
    exec(f.read(), about)


class BaseCommand(Command):
    description = ''
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        pass


class Minimizer(BaseCommand):
    """Minify content of all json file for all locales.
    """

    def initialize_options(self):
        """Find all files of all locales.
        """

        self.paths = []
        self.separators = (',', ':')
        self.data_dir = join(here, 'mimesis', 'data')
        self.before_total = 0
        self.after_total = 0

        for root, _, files in os.walk(self.data_dir):
            for file in sorted(files):
                if splitext(file)[1] == '.json':
                    self.paths.append(join(
                        relpath(root, self.data_dir), file))

    @staticmethod
    def size_of(num):
        for unit in ['B', 'KB', 'MB']:
            if abs(num) < 1024.0:
                return "%3.1f%s" % (num, unit)
            num = num / 1024.0
        return "%.1f" % num

    def minify(self, file_path):
        size_before = getsize(file_path)
        self.before_total += size_before
        size_before = self.size_of(size_before)

        with open(file_path, 'r', 1) as f:
            json_text = json.loads(f.read())
            minimized = json.dumps(
                json_text, separators=self.separators, ensure_ascii=False)

        if len(file_path) > 0:
            output_path = abspath(file_path)
            abs_path = abspath(dirname(output_path))

            if not exists(abs_path):
                os.makedirs(abs_path)

            with open(output_path, 'w+', 1) as f:
                f.write(minimized)

        size_after = getsize(file_path)
        self.after_total += size_after
        size_after = self.size_of(size_after)

        json_file = '/'.join(file_path.split('/')[-2:])

        template = "\033[34m{}\033[0m : " \
                   "\033[92mminimized\033[0m : " \
                   "\033[33m{}\033[0m -> \033[92m{}\033[0m"

        print(template.format(json_file,
                              size_before, size_after))

    def run(self):
        """Start json minimizer and exit when all json
        files was minimized.
        """
        for rel_path in sorted(self.paths):
            file_path = join(self.data_dir, rel_path)
            self.minify(file_path)

        after = self.size_of(self.after_total)
        before = self.size_of(self.before_total)
        saved = self.size_of(self.before_total - self.after_total)

        template = '\nTotal: ' \
                   '\033[92m{}\033[0m -> \033[92m{}\033[0m. ' \
                   'Compressed: \033[92m{}\033[0m\n'

        print(template.format(before, after, saved))


with open('README.rst', 'r', encoding='utf-8') as f:
    readme = f.read()


setup(
    name=about['__title__'],
    version=about['__version__'],
    description=about['__description__'],
    long_description=readme,
    author=about['__author__'],
    author_email=about['__author_email__'],
    url=about['__url__'],
    license=about['__license__'],
    packages=[
        'mimesis',
        'mimesis.data',
        'mimesis.data.int',
        'mimesis.builtins',
        'mimesis.providers',
    ],
    keywords=[
        'fake',
        'mock',
        'data',
        'populate',
        'database',
        'testing',
        'generate',
        'mimesis',
        'dummy',
    ],
    package_data={
        'mimesis': [
            'data/*/*',
        ],
    },
    exclude_package_data={
        'mimesis': [
            # It's for development.
            'data/locale_template/*'
        ],
    },
    data_files=[
        ('', ['LICENSE']),
    ],
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: CPython',
        'Operating System :: OS Independent',
        'Topic :: Software Development',
        'Topic :: Software Development :: Testing',
    ],
    cmdclass={
        'minify': Minimizer,
    },
)



"""Implements classes for generating data by schema."""

from types import LambdaType
from typing import Any, Callable, List, Optional

from mimesis.exceptions import (UndefinedField, UndefinedSchema,
                                UnsupportedField)
from mimesis.providers.base import StrMixin
from mimesis.providers.generic import Generic
from mimesis.typing import JSON, Seed

__all__ = ['AbstractField', 'Field', 'Schema']


class AbstractField(StrMixin):
    """
    AbstractField is a class for generating data by the name of the method.

    Instance of this object takes any string which represents name
    of any method of any supported data provider (:class:`~mimesis.Generic`)
    and the ``**kwargs`` of the method:

    >>> _ = AbstractField('en', 0xf)
    >>> _('full_name')
    'Jack Allison'
    """

    def __init__(self, locale: str = 'en',
                 seed: Optional[Seed] = None,
                 providers: Optional[Any] = None) -> None:
        """Initialize field.

        :param locale: Locale
        :param seed: Seed for random.
        """
        self.locale = locale
        self.seed = seed
        self._gen = Generic(self.locale, self.seed)

        if providers:
            self._gen.add_providers(*providers)

        self._table = {}  # type: ignore

    def __call__(self, name: Optional[str] = None,
                 key: Optional[Callable] = None, **kwargs) -> Any:
        """Override standard call.

        This magic method override standard call so it's take any string which
        represents name of the any method of any supported data provider
        and the ``**kwargs`` of this method.

        ..note:: Some data providers have methods with the same name and
        in such cases, you can explicitly define that the method belongs to
        data-provider ``name='provider.name'``.

        You can apply a *key function* to result returned by the method,
        to do it, just pass parameter **key** with a callable object which
        returns final result.

        :param name: Name of the method.
        :param key: A key function (or other callable object)
            which will be applied to result.
        :param kwargs: Kwargs of method.
        :return: Value which represented by method.
        :raises ValueError: if provider is not
            supported or if field is not defined.
        """
        if name is None:
            raise UndefinedField()

        def tail_parser(tails: str, obj: Any) -> Any:
            """Return method from end of tail.

            :param tails: Tail string
            :param obj: Search tail from this object
            :return last tailed method
            """
            first, second = tails.split('.', 1)
            if hasattr(obj, first):
                attr = getattr(obj, first)
                if '.' in second:
                    return tail_parser(attr, second)
                else:
                    return getattr(attr, second)

        try:
            if name not in self._table:
                if '.' not in name:
                    for provider in dir(self._gen):
                        provider = getattr(self._gen, provider)
                        if name in dir(provider):
                            self._table[name] = getattr(provider, name)
                else:
                    self._table[name] = tail_parser(name, self._gen)

            result = self._table[name](**kwargs)
            if key and callable(key):
                return key(result)
            return result
        except Exception:
            raise UnsupportedField(name)


class Schema(object):
    """Class which return list of filled schemas."""

    def __init__(self, schema: LambdaType) -> None:
        """Initialize schema.

        :param schema: A schema.
        """
        if isinstance(schema, LambdaType):
            self.schema = schema
        else:
            raise UndefinedSchema()

    def create(self, iterations: int = 1) -> List[JSON]:
        """Return filled schema.

        Create a list of a filled schemas with elements in
        an amount of **iterations**.

        :param iterations: Amount of iterations.
        :return: List of willed schemas.
        """
        return [self.schema() for _ in range(iterations)]


# Alias for AbstractField
Field = AbstractField



"""Provides all the data related to the development."""

CONTAINER = [
    'Docker',
    'Rkt',
    'LXC',
    'Vagrant',
    'OpenVZ',
    'Kubernetes',
]

VERSION_CONTROL_SYSTEMS = [
    'Git',
    'Mercurial',
    'Subversion',
]

LICENSES = [
    'Apache License, 2.0 (Apache-2.0)',
    'The BSD 3-Clause License',
    'The BSD 2-Clause License',
    'GNU General Public License (GPL)',
    'General Public License (LGPL)',
    'MIT License (MIT)',
    'Mozilla Public License 2.0 (MPL-2.0)',
    'Common Development and Distribution License (CDDL-1.0)',
    'Eclipse Public License (EPL-1.0)',
]

PROGRAMMING_LANGS = [
    'ASP',
    'Assembly',
    'AutoIt',
    'Awk',
    'Bash',
    'C',
    'C Shell',
    'C#',
    'C++',
    'Caml',
    'Ceylon',
    'Clojure',
    'CoffeeScript',
    'Common Lisp',
    'D',
    'Dart',
    'Delphi',
    'Dylan',
    'ECMAScript',
    'Elixir',
    'Emacs Lisp',
    'Erlang',
    'F#',
    'Falcon',
    'Fortran',
    'GNU Octave',
    'Go',
    'Groovy',
    'Haskell',
    'haXe',
    'Io',
    'J#',
    'Java',
    'JavaScript',
    'Julia',
    'Kotlin',
    'Lisp',
    'Lua',
    'Mathematica',
    'Objective-C',
    'OCaml',
    'Perl',
    'PHP',
    'PL-I',
    'PL-SQL',
    'PowerShell',
    'Prolog',
    'Python',
    'R',
    'Racket',
    'Ruby',
    'Rust',
    'Scala',
    'Scheme',
    'Smalltalk',
    'Tcl',
    'Tex',
    'Transact-SQL',
    'TypeScript',
    'Z shell',
]

NOSQL = [
    'Aerospike',
    'Aerospike',
    'AllegroGraph',
    'Apache Cassandra',
    'CouchDB',
    'Couchbase',
    'Hyperdex',
    'InfiniteGraph',
    'MUMPS',
    'MemcacheDB',
    'Memcached',
    'MongoDB',
    'Neo4J',
    'Redis',
    'RethinkDB',
    'Riak',
]

SQL = [
    'MariaDB',
    'MySQL',
    'PostgreSQL',
    'Oracle DB',
    'SQLite',
    'DB2',
]

BACKEND = [
    'C#/ASP.NET',
    'Java',
    'Java/Tapestry',
    'Go/Beego',
    'Python/Bottle',
    'Python/CherryPy',
    'Erlang',
    'Erlang/ChicagoBoss',
    'Erlang/Cowboy',
    'Clojure',
    'Django',
    'Python/Django',
    'Python/Flask/Gunicorn',
    'Elixir/Phoenix',
    'Node/Express',
    'Python/Falcon',
    'Python/Flask',
    'Go/Gin',
    'Go/Martini',
    'Go/Gorilla',
    'Groovy/Grails',
    'PHP/Laravel',
    'Java/Ninja',
    'PHP/Yii 2',
    'Pyramid',
    'Python/Pyramid',
    'Python/SQLAlchemy',
    'Python/Twisted',
    'Python/web2py',
    'Go/Revel',
    'Ruby on Rails',
    'Ruby/Sinatra',
    'Java/Spring',
    'PHP/Symfony',
    'Python/Tornado',
    'PHP/Zend Framework',
]

FRONTEND = [
    'JavaScript/Angular 2',
    'JavaScript/Backbone',
    'Gulp/Bower/Sass',
    'JS/HTML/CSS/Canvas/SVG',
    'JavaScript/Ember',
    'JavaScript/React',
    'JavaScript/React/Redux',
    'JavaScript/React/Redux/Flux',
    'JS/HTML/CSS',
    'JavaScript/Highcharts.js',
    'JS/Twitter Bootstrap',
    'React/Canvas/SVG',
    'React/D3.js',
    'React/Redux/Babel',
    'JavaScript/Reflux/Redux',
    'JS/Sass/Less/Gulp',
    'Stylus/Sass/Webpack',
    'JavaScript/Vue',
    'JavaScript/jQuery',
    'TypeScript/React',
    'TypeScript/CSS/HTML',
    'Webpack/Grunt/PostCSS',
]

OS = [
    'Arch',
    'CentOS',
    'Debian',
    'Fedora',
    'FreeBSD',
    'Gentoo',
    'Kali',
    'Lubuntu',
    'Manjaro',
    'Mint',
    'OS X',
    'OpenBSD',
    'PCLinuxOS',
    'Slackware',
    'Ubuntu',
    'Windows 10',
    'Windows 7',
    'Windows 8',
    'Windows 8.1',
    'Zorin',
    'elementaryOS',
    'macOS',
    'openSUSE',
]

FOLDERS = [
    'Development',
    'Downloads',
    'Documents',
    'Music',
    'Video',
    'Work',
    'Pictures',
    'Desktop',
    'Study',
]

PROJECT_NAMES = [
    'aardonyx',
    'abelisaurus',
    'achelousaurus',
    'achillobator',
    'acrocanthosaurus',
    'aegyptosaurus',
    'afrovenator',
    'agilisaurus',
    'alamosaurus',
    'albertaceratops',
    'albertosaurus',
    'alectrosaurus',
    'alioramus',
    'allosaurus',
    'alvarezsaurus',
    'amargasaurus',
    'ammosaurus',
    'ampelosaurus',
    'amygdalodon',
    'anatotitan',
    'anchiceratops',
    'anchisaurus',
    'ankylosaurus',
    'anserimimus',
    'antarctopelta',
    'antarctosaurus',
    'apatosaurus',
    'aragosaurus',
    'aralosaurus',
    'archaeoceratops',
    'archaeopteryx',
    'archaeornithomimus',
    'argentinosaurus',
    'arrhinoceratops',
    'atlascopcosaurus',
    'aucasaurus',
    'austrosaurus',
    'avaceratops',
    'avalonia',
    'avimimus',
    'azendohsaurus',
    'bactrosaurus',
    'bagaceratops',
    'bambiraptor',
    'barapasaurus',
    'barosaurus',
    'baryonyx',
    'becklespinax',
    'beipiaosaurus',
    'bellusaurus',
    'borogovia',
    'brachiosaurus',
    'brachyceratops',
    'bugenasaura',
    'buitreraptor',
    'camarasaurus',
    'camptosaurus',
    'carnotaurus',
    'caudipteryx',
    'cedarpelta',
    'centrosaurus',
    'ceratosaurus',
    'cetiosauriscus',
    'cetiosaurus',
    'chaoyangsaurus',
    'chasmosaurus',
    'chialingosaurus',
    'chindesaurus',
    'chinshakiangosaurus',
    'chirostenotes',
    'chubutisaurus',
    'chungkingosaurus',
    'citipati',
    'coelophysis',
    'coelurus',
    'coloradisaurus',
    'compsognathus',
    'conchoraptor',
    'confuciusornis',
    'corythosaurus',
    'cryolophosaurus',
    'dacentrurus',
    'daspletosaurus',
    'datousaurus',
    'deinocheirus',
    'deinonychus',
    'deltadromeus',
    'diceratops',
    'dicraeosaurus',
    'dilophosaurus',
    'diplodocus',
    'dracorex',
    'dravidosaurus',
    'dromaeosaurus',
    'dromiceiomimus',
    'dryosaurus',
    'dryptosaurus',
    'dubreuillosaurus',
    'edmontonia',
    'edmontosaurus',
    'einiosaurus',
    'elaphrosaurus',
    'emausaurus',
    'eolambia',
    'eoraptor',
    'eotyrannus',
    'equijubus',
    'erketu',
    'erlikosaurus',
    'euhelopus',
    'euoplocephalus',
    'europasaurus',
    'euskelosaurus',
    'eustreptospondylus',
    'fukuiraptor',
    'fukuisaurus',
    'gallimimus',
    'gargoyleosaurus',
    'garudimimus',
    'gasosaurus',
    'gasparinisaura',
    'gastonia',
    'giganotosaurus',
    'gilmoreosaurus',
    'giraffatitan',
    'gobisaurus',
    'gorgosaurus',
    'goyocephale',
    'graciliceratops',
    'gryposaurus',
    'guaibasaurus',
    'guanlong',
    'hadrosaurus',
    'hagryphus',
    'haplocanthosaurus',
    'harpymimus',
    'herrerasaurus',
    'hesperosaurus',
    'heterodontosaurus',
    'homalocephale',
    'huayangosaurus',
    'hylaeosaurus',
    'hypacrosaurus',
    'hypselosaurus',
    'hypsilophodon',
    'iguanodon',
    'indosuchus',
    'ingenia',
    'irritator',
    'isisaurus',
    'janenschia',
    'jaxartosaurus',
    'jingshanosaurus',
    'jinzhousaurus',
    'jobaria',
    'juravenator',
    'kentrosaurus',
    'khaan',
    'kotasaurus',
    'kritosaurus',
    'lamaceratops',
    'lambeosaurus',
    'lapparentosaurus',
    'leaellynasaura',
    'leptoceratops',
    'lesothosaurus',
    'lexovisaurus',
    'liaoceratops',
    'liaoxiornis',
    'ligabuesaurus',
    'liliensternus',
    'lophorhothon',
    'lophostropheus',
    'lufengosaurus',
    'lurdusaurus',
    'lycorhinus',
    'magyarosaurus',
    'maiasaura',
    'majungatholus',
    'malawisaurus',
    'mamenchisaurus',
    'mapusaurus',
    'marshosaurus',
    'masiakasaurus',
    'massospondylus',
    'maxakalisaurus',
    'megalosaurus',
    'melanorosaurus',
    'metriacanthosaurus',
    'microceratops',
    'micropachycephalosaurus',
    'microraptor',
    'minmi',
    'monolophosaurus',
    'mononykus',
    'mussaurus',
    'muttaburrasaurus',
    'nanotyrannus',
    'nanshiungosaurus',
    'nemegtosaurus',
    'neovenator',
    'neuquenosaurus',
    'nigersaurus',
    'nipponosaurus',
    'noasaurus',
    'nodosaurus',
    'nomingia',
    'nothronychus',
    'nqwebasaurus',
    'omeisaurus',
    'ornitholestes',
    'ornithomimus',
    'orodromeus',
    'oryctodromeus',
    'othnielia',
    'ouranosaurus',
    'oviraptor',
    'rebbachisaurus',
    'rhabdodon',
    'rhoetosaurus',
    'rinchenia',
    'riojasaurus',
    'rugops',
    'saichania',
    'saltasaurus',
    'saltopus',
    'sarcosaurus',
    'saurolophus',
    'sauropelta',
    'saurophaganax',
    'saurornithoides',
    'scelidosaurus',
    'scutellosaurus',
    'secernosaurus',
    'segisaurus',
    'segnosaurus',
    'seismosaurus',
    'shamosaurus',
    'shanag',
    'shantungosaurus',
    'shunosaurus',
    'shuvuuia',
    'silvisaurus',
    'sinocalliopteryx',
    'sinornithosaurus',
    'sinosauropteryx',
    'sinraptor',
    'sinvenator',
    'zalmoxes',
    'zephyrosaurus',
    'zuniceratops',
    'byzantine',
    'svengali',
    'accolade',
    'acrimony',
    'angst',
    'anomaly',
    'antidote',
    'baroque',
    'bona_fide',
    'bourgeois',
    'bravado',
    'brogue',
    'brusque',
    'cacophony',
    'caustic',
    'charisma',
    'cloying',
    'deja-vu',
    'dichotomy',
    'elan',
    'ennui',
    'epitome',
    'esoteric',
    'euphemism',
    'faux pas',
    'fiasco',
    'finagle',
    'glib',
    'harbinger',
    'hedonist',
    'heresy',
    'idyllic',
    'insidious',
    'junket',
    'kitsch',
    'litany',
    'lurid',
    'malaise',
    'malinger',
    'mantra',
    'maudlin',
    'mercenary',
    'misnomer',
    'nirvana',
    'oblivion',
    'ogle',
    'ostracize',
    'panacea',
    'paradox',
    'peevish',
    'propriety',
    'revel',
    'rhetoric',
    'spartan',
    'stigma',
    'stoic',
    'suave',
    'sycophant',
    'tirade',
    'tryst',
    'untenable',
    'vicarious',
    'vile',
    'waft',
    'zealous',
]



"""Provides data related to gaming and games."""

from mimesis.data import GAMES, GAMING_PLATFORMS, GENRES, SCORE_PHRASES
from mimesis.providers.base import BaseDataProvider

__all__ = ['Games']


class Games(BaseDataProvider):
    """Class for generating data related to the games."""

    def gaming_platform(self) -> str:
        """Get random gaming platform.

        :return: Gaming platform

        :Example:
            PlayStation 4 Pro
        """
        return self.random.choice(GAMING_PLATFORMS)

    def score(self, minimum: int = 1, maximum: int = 10) -> float:
        """Score of game.

        :param minimum: Maximum value.
        :param maximum: Minimum value.
        :return: Score.
        """
        return self.random.randint(minimum * 10, maximum * 10) / 10

    def pegi_rating(self) -> str:
        """Get a random PEGI rating.

        :return: PEGI rating.

        :Example:
            PEGI 18
        """
        standard = [3, 7, 12, 16, 18]
        digit = self.random.choice(standard)
        return '{0} {1}'.format('PEGI', digit)

    def genre(self) -> str:
        """Get a random genre of game.

        :return: Genre.

        :Example:
            Shooter
        """
        return self.random.choice(GENRES)

    def score_phrase(self) -> str:
        """Get a random score phrase.

        :return: Score phrase.

        :Example:
            Great
        """
        return self.random.choice(SCORE_PHRASES)

    def game(self) -> str:
        """Get a random game from list of games.

        :return: Name of the game.

        :Example:
            Battlefield 1
        """
        return self.random.choice(GAMES)



"""Specific data provider for Brazil (pt-br)."""

from random import randint

from mimesis.builtins.base import BaseSpecProvider

__all__ = ['BrazilSpecProvider']


class BrazilSpecProvider(BaseSpecProvider):
    """Class that provides special data for Brazil (pt-br)."""

    class Meta:
        """The name of the provider."""

        name = 'brazil_provider'

    @staticmethod
    def cpf(with_mask: bool = True) -> str:
        """Get a random CPF.

        :param with_mask: Use CPF mask (###.###.###-##).
        :returns: Random CPF.

        :Example:
            001.137.297-40
        """
        def get_verifying_digit_cpf(cpf, peso):
            """Calculate the verifying digit for the CPF.

            :param cpf: List of integers with the CPF.
            :param peso: Integer with the weight for the modulo 11 calculate.
            :returns: The verifying digit for the CPF.
            """
            soma = 0
            for index, digit in enumerate(cpf):
                soma += digit * (peso - index)
            resto = soma % 11
            if resto == 0 or resto == 1 or resto >= 11:
                return 0
            return 11 - resto

        cpf_without_dv = [randint(0, 9) for _ in range(9)]
        first_dv = get_verifying_digit_cpf(cpf_without_dv, 10)

        cpf_without_dv.append(first_dv)
        second_dv = get_verifying_digit_cpf(cpf_without_dv, 11)
        cpf_without_dv.append(second_dv)

        cpf = ''.join([str(i) for i in cpf_without_dv])

        if with_mask:
            return cpf[:3] + '.' + cpf[3:6] + '.' + cpf[6:9] + '-' + cpf[9:]
        return cpf

    @staticmethod
    def cnpj(with_mask: bool = True) -> str:
        """Get a random CNPJ.

        :param with_mask: Use cnpj mask (###.###.###-##)
        :returns: Random cnpj.

        :Example:
            77.732.230/0001-70
        """
        def get_verifying_digit_cnpj(cnpj, peso):
            """Calculate the verifying digit for the CNPJ.

            :param cnpj: List of integers with the CNPJ.
            :param peso: Integer with the weight for the modulo 11 calculate.
            :returns: The verifying digit for the CNPJ.
            """
            soma = 0
            if peso == 5:
                peso_list = [5, 4, 3, 2, 9, 8, 7, 6, 5, 4, 3, 2]
            elif peso == 6:
                peso_list = [6, 5, 4, 3, 2, 9, 8, 7, 6, 5, 4, 3, 2]
            for i, _ in enumerate(cnpj):
                soma += peso_list[i] * cnpj[i]
            resto = soma % 11
            if resto < 2:
                return 0
            return 11 - resto

        cnpj_without_dv = [randint(0, 9) for _ in range(12)]

        first_dv = get_verifying_digit_cnpj(cnpj_without_dv, 5)
        cnpj_without_dv.append(first_dv)

        second_dv = get_verifying_digit_cnpj(cnpj_without_dv, 6)
        cnpj_without_dv.append(second_dv)

        cnpj = ''.join([str(i) for i in cnpj_without_dv])

        if with_mask:
            return (
                cnpj[:2] + '.' + cnpj[2:5] + '.' + cnpj[5:8] + '/' +
                cnpj[8:12] + '-' + cnpj[12:]
            )
        return cnpj



"""Data related to the development."""

from mimesis.data import (BACKEND, CONTAINER, FRONTEND, LICENSES, NOSQL, OS,
                          PROGRAMMING_LANGS, SQL, VERSION_CONTROL_SYSTEMS)
from mimesis.providers.base import BaseDataProvider

__all__ = ['Development']


class Development(BaseDataProvider):
    """Class for getting fake data for Developers."""

    def software_license(self) -> str:
        """Get a random software license.

        :return: License name.

        :Example:
            The BSD 3-Clause License.
        """
        return self.random.choice(LICENSES)

    def version(self, calver: bool = False, pre_release: bool = False) -> str:
        """Generate version number.

        :param calver: Calendar versioning.
        :param pre_release: Pre-release.
        :return: Version.

        :Example:
            0.2.1
        """
        # TODO: Optimize
        version = '{}.{}.{}'
        major, minor, patch = self.random.randints(3, 0, 10)

        if calver:
            if minor == 0:
                minor += 1

            if patch == 0:
                patch += 1
            major = self.random.randint(2016, 2018)
            return version.format(major, minor, patch)

        version = '{}.{}.{}'.format(major, minor, patch)

        if pre_release:
            suffixes = ('alpha', 'beta', 'rc')
            suffix = self.random.choice(suffixes)
            number = self.random.randint(1, 11)
            return '{}-{}.{}'.format(version, suffix, number)

        return version

    def database(self, nosql: bool = False) -> str:
        """Get a random database name.

        :param nosql: only NoSQL databases.
        :return: Database name.

        :Example:
            PostgreSQL.
        """
        if nosql:
            return self.random.choice(NOSQL)
        return self.random.choice(SQL)

    def container(self) -> str:
        """Get a random containerization system.

        :return: Containerization system.

        :Example:
            Docker.
        """
        return self.random.choice(CONTAINER)

    def version_control_system(self) -> str:
        """Get a random version control system.

        :return: Version control system

        :Example:
            Git
        """
        return self.random.choice(VERSION_CONTROL_SYSTEMS)

    def programming_language(self) -> str:
        """Get a random programming language from the list.

        :return: Programming language.

        :Example:
            Erlang.
        """
        return self.random.choice(PROGRAMMING_LANGS)

    def backend(self) -> str:
        """Get a random backend stack.

        :return: Stack.

        :Example:
            Elixir/Phoenix
        """
        return self.random.choice(BACKEND)

    def frontend(self) -> str:
        """Get a random front-end stack.

        :return: Stack.

        :Example:
            JS/React.
        """
        return self.random.choice(FRONTEND)

    def os(self) -> str:
        """Get a random operating system or distributive name.

        :return: The name of OS.

        :Example:
            Gentoo
        """
        return self.random.choice(OS)

    def boolean(self) -> bool:
        """Get a random boolean value.

        :return: True of False.
        """
        return self.random.choice([True, False])



"""Decorators for the public API and for internal purpose."""

import functools
from string import ascii_letters as letters
from string import digits, punctuation
from typing import Callable

from mimesis import data
from mimesis.exceptions import UnsupportedLocale


def romanized(locale: str = '') -> Callable:
    """Romanize the Cyrillic text.

    Transliterate the Cyrillic language from the Cyrillic
    script into the Latin alphabet.

    .. note:: At this moment it works only for `ru`, `uk`, `kk`.

    :param locale: Locale code.
    :return: Latinized text.
    """
    def romanized_deco(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            try:
                alphabet = data.ROMANIZATION_DICT[locale]
                # Add common cyrillic common letters
                alphabet.update(data.COMMON_LETTERS)
                # String can contain ascii symbols, digits and
                # punctuation symbols.
                alphabet.update({s: s for s in
                                 letters + digits + punctuation})
            except KeyError:
                raise UnsupportedLocale(locale)
            result = func(*args, **kwargs)
            txt = ''.join([alphabet[i] for i in result if i in alphabet])
            return txt

        return wrapper

    return romanized_deco
